/* --------------------------------------------------------------------------
 *
 *  WP
 *
 * --------------------------------------------------------------------------*/

#include "wpj.h"

typedef struct io_connect_st
{
    pool p;
    char *ip;
    int port;
    void *cb;
    void *cb_arg;
    pthread_t t;
    int timeout;
    int connected;
} _connect_data,  *connect_data;


/* global object */
ios io__data = NULL;
extern wpjs wpj;

#define KARMA_HEARTBEAT_LK 2  /* 2 sec */
int KARMA_DEF_INIT    = KARMA_INIT;
int KARMA_DEF_MAX     = KARMA_MAX;
int KARMA_DEF_INC     = KARMA_INC;
int KARMA_DEF_DEC     = KARMA_DEC;
int KARMA_DEF_PENALTY = KARMA_PENALTY;
int KARMA_DEF_RESTORE = KARMA_RESTORE;

/* RATE HASHTBALE */
int DEF_RATE_TIME  = 1; /* 2 connection on 1 second */
int DEF_RATE_CONN  = 2;
int DEF_CONN  = 0; /* no limit */


void update_links_count(){
  FILE * plik;

    plik = fopen(wpj->cfg->statfile, "w");
    if (!plik) plik = fopen("/tmp/liczba_polaczen", "w");
    if (plik) {
      fprintf(plik, "%d", wpj->clients_count);
      fclose(plik);
    }
}

void karma_decrement_lk(struct karma *k)
{
    /* lower our karma */
    k->val -= k->dec;

    /* if below zero, set to penalty */
    if( k->val <= 0 ) 
        k->val = k->penalty;
}

/*
 * returns 1 - k.val > 0;
 */
int karma_increment_lk(struct karma *k, DWORD time)
{
  int punishment_over = 0;
  DWORD dt;

  /* only increment every KARMA_HEARTBEAT seconds */
  if( ( k->last_update + KARMA_HEARTBEAT_LK > time ) && k->last_update != 0)
    goto CHECK;

  /* do not raise karma if overload */
  if (wpj->overload){
    k->last_update = time;   
    goto CHECK;
  }
  
  dt = time - k->last_update; /* might be HEARBEAT, HEARTBEAT+1 .... */
  dt = dt / KARMA_HEARTBEAT_LK;
  
  /* if incrementing will raise over 0 */
  if( ( k->val < 0 ) && ( k->val + (k->inc * dt) > 0 ) )
    punishment_over = 1;
  
  /* increment the karma value */
  k->val += (k->inc * dt);
  if( k->val > k->max ) k->val = k->max; /* can only be so good */
  
  /* lower our byte count, if we have good karma */
  if( k->val > 0 ) k->bytes -= ( KARMA_READ_MAX(k->val) );
  if( k->bytes <0 ) k->bytes = 0;
  
  k->last_update = time;

  log_debug(ZONE,"karma increment to %d o %d",k->val,dt);

  /* our karma has *raised* to 0 */
  if( punishment_over )
    {
      k->val = k->restore;
      return 1; /* k.val > 0 */
    }
  
 CHECK:
  if (k->val > 0)
    return 1;
  else
    return 0;
}

int karma_check_lk(struct karma *k,long bytes_read)
{
    /* add up the total bytes */
    k->bytes += bytes_read;
    if( k->bytes > KARMA_READ_MAX(k->val) )
        karma_decrement_lk( k );

    /* check if it's okay */
    if( k->val <= 0 )
        return 1;

    /* everything is okay */
    return 0;
}

#define CLIENT_HASH_SIZE 60000

io client_hash[CLIENT_HASH_SIZE]; /* contents info about users fd -> m */

/* hash */
void remove_client(int fd)
{
  client_hash[fd]=NULL;
}

void add_client(int fd,io pos)
{
  if (client_hash[fd] != NULL){
    log_alert(ZONE,"error in client hash, overwrite");
  }
    
  client_hash[fd] = pos;
}

io find_client(int fd)
{
  log_debug(ZONE,"find client %d",fd);
  return client_hash[fd]; 
}

/* 
 * Definitions
 */
int _io_write_dump_server(io m, int * length);

/* 
 * Check socket for writeing
 */
int IOCanWrite(int fd)
{
    struct timeval tv;
    fd_set set;

    tv.tv_sec=0;
    tv.tv_usec=0;
    FD_ZERO(&set);
    
    FD_SET(fd,&set);
    return select(fd+1,NULL,&set,NULL,&tv);
}

/* 
 * Server writing queue thread
 */
void * write_queue_main(void *arg)
{
  io cur;
  int ret;
  DWORD time;
  int length;

  log_debug(ZONE,"Write thread started");
    
  while (1)
    {
        /* if we are closing down, exit the loop */
        if (io__data->shutdown == 3)
	{
	   break;
	}

	/* sleep 1/1000  - 1000 writes per seconds */
	Sleep(1);

	/* time */
	time = timeGetTimeSec();

	/* write only to authorised servers */
	cur = io__data->serverio;
	if (cur) /* if server */
	  if (cur->state == state_ACTIVE) /* if active */
	    if (cur->queue != NULL)  /* if data */
            {   
	      /* write the current buffer */
	      ret = _io_write_dump_server(cur,&length);
	      if (ret == -1) {
		  log_alert(ZONE,"Blad pisania do serwera");
		  io_close(cur);
	      }
	      else {
		if (cur->ping_in_progress) {
		  cur->write_buffer_count += length;
		}
	      }

		/* check overload */
	        if (wpj->overload)/* 1 or 2 */
		{		  
		  if (cur->queue_long < OVERLOAD_RET)
		  {
			wpj->overload = 0;
			log_alert("--","0");
		  }
		  else
		  if ((wpj->overload == 2)&&(cur->queue_long < OVERLOAD_NORMAL))
		  {		  		  
			wpj->overload = 1;
			log_alert("--","1");
		  }
		}
		
	    }
    }
    
    return 0;
}

/* 
 * dev/epoll set 
 */
int fdop(int fd,int op){
  struct pollfd pfd;
  pfd.fd = fd;
  pfd.events = op;
  pfd.revents = 0;
  if (write(io__data->epollfd, &pfd, sizeof(pfd)) != sizeof(pfd)) {
    log_alert(ZONE,"fdop error");
    return -1;
  }
  return 0;
}
#define ADDFD POLLIN | POLLOUT | POLLERR | POLLHUP
#define REMFD POLLREMOVE

/* 
 * unlinks a socket from the master list 
 */
void _io_unlink(io m)
{
    if(io__data == NULL) 
        return;

    io__data->list_count--;

    if(io__data->master__list == m)
       io__data->master__list = io__data->master__list->next;

    if(m->prev != NULL) 
        m->prev->next = m->next;

    if(m->next != NULL) 
        m->next->prev = m->prev;
}

void _io_unlink_data(io m)
{
    if(io__data == NULL) 
        return;

    if(io__data->data__list == m)
       io__data->data__list = io__data->data__list->datanext;

    if(m->dataprev != NULL) 
        m->dataprev->datanext = m->datanext;

    if(m->datanext != NULL) 
        m->datanext->dataprev = m->dataprev;
}

/* 
 * links a socket to the master list 
 */
void _io_link(io m)
{
    if(io__data == NULL) 
        return;

    io__data->list_count++;

    m->next = io__data->master__list;
    m->prev = NULL;

    if(io__data->master__list != NULL) 
      io__data->master__list->prev = m;

    io__data->master__list = m;
}

void _io_link_data(io m)
{
    if(io__data == NULL) 
      return;
    
    m->datanext = io__data->data__list;
    m->dataprev = NULL;
    
    if(io__data->data__list != NULL) 
      io__data->data__list->dataprev = m;
    
    io__data->data__list = m;
}

/* 
 * Dump this socket's write queue.  tries to write
 * as much of the write queue as it can, before the
 * write call would block the server
 * returns -1 on error, 0 on success, and 1 if more data to write
 */
int _io_write_dump_client(io m)
{
    int len;
    io_wbq cur;

    /* try to write as much as we can */
    while(m->queue != NULL)
    {
        cur = m->queue;

        log_debug(ZONE, "write_dump client writing data: %.*s", cur->len, cur->cur);

        /* write a bit from the current buffer */
        len = IO_WRITE_FUNC(m->fd, cur->cur, cur->len);

        /* we had an error on the write */
        if(len == 0)
        {
	  log_alert(ZONE,"error in writing, len=0");
            if(m->cb != NULL)
                (*(io_std_cb)m->cb)(m, IO_ERROR, m->cb_arg);
				
            return -1;
        }
        if(len < 0)
        { 
            /* if we have an error, that isn't a blocking issue */ 
            if(errno != EWOULDBLOCK && errno != EINTR && errno != EAGAIN)
            { 
	      log_alert(ZONE,"error in writing, len < 0");
                /* bounce the queue */
                if(m->cb != NULL)
                    (*(io_std_cb)m->cb)(m, IO_ERROR, m->cb_arg);

                return -1;
            }
            return 1;
        }
        /* we didnt' write it all, move the current buffer up */
        else if(len < cur->len)
        { 

            cur->cur += len;
            cur->len -= len;
            return 1;
        } 
        /* we wrote the entire node, kill it and move on */
        else
        {  
            m->queue = m->queue->next;

            if(m->queue == NULL)
                m->tail = NULL;

            pool_free(cur->p);
	    m->queue_long--;	
        }
    } 
    return 0;
}


int _io_write_dump_server(io m, int * length)
{
    int len;
    io_wbq cur;

    length = 0;

    pthread_mutex_lock(&(m->sem));

    /* try to write as much as we can */
    while(m->queue != NULL)
    {
        cur = m->queue;

        log_debug(ZONE, "write_dump serwer writing data: %.*s", cur->len, cur->cur);

        /* write a bit from the current buffer */
        len = IO_WRITE_FUNC(m->fd, cur->cur, cur->len);

        /* we had an error on the write */
        if(len == 0)
        {
	    pthread_mutex_unlock(&(m->sem));
	
            if(m->cb != NULL)
                (*(io_std_cb)m->cb)(m, IO_ERROR, m->cb_arg);
				
            return -1;
        }
        if(len < 0)
        { 
            /* if we have an error, that isn't a blocking issue */ 
            if(errno != EWOULDBLOCK && errno != EINTR && errno != EAGAIN)
            { 
		pthread_mutex_unlock(&(m->sem));

		log_error(ZONE,"Write ERROR !!!"); /* sielim PATCH */
			    
                /* bounce the queue */
                if(m->cb != NULL)
                    (*(io_std_cb)m->cb)(m, IO_ERROR, m->cb_arg);

                return -1;
            }
	    pthread_mutex_unlock(&(m->sem));

	    /* didn't write anything */
            return 1;
        }
        /* we didnt' write it all, move the current buffer up */
        else if(len < cur->len)
	{
	    cur->cur += len;
            cur->len -= len;
	    pthread_mutex_unlock(&(m->sem));
	    length += len;
            return 1;
        } 
        /* we wrote the entire node, kill it and move on */
        else
        {  
  	    length += len;

            m->queue = m->queue->next;

            if(m->queue == NULL)
                m->tail = NULL;

            pool_free(cur->p);
	    m->queue_long--;	
        }
    } 

    pthread_mutex_unlock(&(m->sem));
    return 0;
}

/* 
 * internal close function 
 * does a final write of the queue, bouncing and freeing all memory
 */
void _io_close(io m)
{
    int ret = 0;
    xmlnode cur;

    log_debug(ZONE,"_io_close"); 

    /* ensure that the state is set to CLOSED */
    m->state = state_CLOSE;

    RateRemove(m->sin_addr);

    /* take it off the master__list */
    _io_unlink(m);
    if (m->is_data) _io_unlink_data(m);

    /* try to write what's in the queue */
    if(m->queue != NULL) 
    {
        log_debug(ZONE,"dumping data in _io_close");  /* sielim temp */
        ret = _io_write_dump_client(m);
    }

    if(ret == 1) /* still more data, bounce it all */
        if(m->cb != NULL)
            (*(io_std_cb)m->cb)(m, IO_ERROR, m->cb_arg);

    /* notify of the close */
    if(m->cb != NULL)
        (*(io_std_cb)m->cb)(m, IO_CLOSED, m->cb_arg);

    /* close the socket, and free all memory */
    close(m->fd);

    /* cleanup the write queue */
    while((cur = io_cleanup(m)) != NULL)
        xmlnode_free(cur);

    pool_free(m->p);

    log_alert(ZONE,"freed socket");
}


void _io_close_server(io m)
{
    int ret;
    xmlnode cur;
    int length;

    /* ensure that the state is set to CLOSED */
    m->state = state_CLOSE;

    /* try to write what's in the queue */
    if(m->queue != NULL) 
    {
        ret = _io_write_dump_server(m,&length);
    }

    if(ret == 1) /* still more data, bounce it all */
        if(m->cb != NULL)
            (*(io_std_cb)m->cb)(m, IO_ERROR, m->cb_arg);

    /* notify of the close */
    if(m->cb != NULL)
        (*(io_std_cb)m->cb)(m, IO_CLOSED, m->cb_arg);

    /* close the socket, and free all memory */
    io__data->serverio = NULL;

    close(m->fd);

    /* cleanup the write queue */
    while((cur = io_cleanup(m)) != NULL)
        xmlnode_free(cur);

    pthread_mutex_destroy(&(m->sem));

    pool_free(m->p);

    log_alert(ZONE,"freed serwer IO socket");
}

void close_client(io m)
{
  fdop(m->fd,REMFD);
  remove_client(m->fd);
  _io_close(m);
}


void _io_close_listen(io m)
{
    if (m != io__data->listenio) return;

    /* add to dev/epoll */
    fdop(m->fd, REMFD);

    /* close the socket, and free all memory */
    close(m->fd);

    pool_free(m->p);

    io__data->listenio = NULL;

    log_debug(ZONE,"freed listen IO socket");
}


/* 
 * accept an incoming connection from a listen sock 
 */
io _io_accept(io m)
{
    struct sockaddr_in serv_addr;
    size_t addrlen = sizeof(serv_addr);
    int fd;
    io new;
    int s = 4;
    int socket_size = 32768;

    log_debug(ZONE, "_io_accept calling accept on fd #%d", m->fd);

    /* pull a socket off the accept queue */
    fd = IO_ACCEPT_FUNC(m->fd, (struct sockaddr*)&serv_addr, &addrlen);
    if(fd <= 0)
    {
        return ((void *)-1);
    }

    if( wpj->clients_count >= wpj->cfg->max_clients )
    {
        log_warn("accept", "Max clients reached");
        close(fd);
        return NULL;
    }

    /* make sure that we aren't rate limiting this IP */
    if (RateCheckAdd(serv_addr.sin_addr))
    {
        log_warn("io_select", "%s is being connection rate limited", inet_ntoa(serv_addr.sin_addr));
        close(fd);
        return NULL;
    }

    /* set write bufor to 32768 */
    setsockopt(fd,SOL_SOCKET,SO_SNDBUF,(char *)&socket_size,s);


    log_debug(ZONE, "new socket accepted (fd: %d, ip: %s, port: %d)", fd, inet_ntoa(serv_addr.sin_addr), ntohs(serv_addr.sin_port));

    /* create a new sock object for this connection */
    new      = io_new(fd, m->cb, m->cb_arg);
    new->ip  = pstrdup(new->p, inet_ntoa(serv_addr.sin_addr));
    new->sin_addr = serv_addr.sin_addr;

    io_karma2(new, &m->k);

    if(m->cb != NULL)
        (*(io_std_cb)new->cb)(new, IO_NEW, new->cb_arg);

    return new;
}

int poll(struct pollfd *ufds, unsigned int nfds, int timeout);

void * _io_connect(void *arg)
{
    connect_data cd = (connect_data)arg;
    struct sockaddr_in sa;
    struct in_addr     *saddr;
    int                flag = 1,
                       flags;
    int                n;
    io                 new;
    pool               p;
    struct pollfd pfdset[1];
    int                ret = 0;
    int                error,s=4;
    int                socket_size=32768;


    /* wait 2 seconds */
    if (cd->timeout > 0)
      sleep(cd->timeout);

    if (io__data->shutdown > 0)  {
      pool_free(cd->p);
      return (void*)1;
    }

    bzero((void*)&sa, sizeof(struct sockaddr_in));
  
    p           = cd->p;
    new         = pmalloco(p, sizeof(_io));
    new->p      = p;
    new->type   = type_SERVER;
    new->state  = state_ACTIVE;
    new->ip     = pstrdup(p,cd->ip);
    new->cb     = (void*)cd->cb;
    new->cb_arg = cd->cb_arg;
    new->pollpos= -1;
  
    new->fd = socket(AF_INET, SOCK_STREAM,0);

    log_debug(ZONE,"Connecting server on socket %d",new->fd);
  
    if(new->fd < 0 || setsockopt(new->fd, SOL_SOCKET, SO_REUSEADDR, (char*)&flag, sizeof(flag)) < 0)
    {
        if(cd->cb != NULL)
            (*(io_std_cb)cd->cb)(new, IO_CLOSED, cd->cb_arg);
        cd->connected = -1;

        if(new->fd > 0)
            close(new->fd);
        pool_free(p);
	return (void*)1;
    }

    flags =  fcntl(new->fd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    fcntl(new->fd, F_SETFL, flags);

    saddr = make_addr(cd->ip);
    if(saddr == NULL)
    {
        if(cd->cb != NULL)
            (*(io_std_cb)cd->cb)(new, IO_CLOSED, cd->cb_arg);
        cd->connected = -1;

        if(new->fd > 0)
            close(new->fd);
        pool_free(p);
	return (void*)1;
    }

    sa.sin_family = AF_INET;
    sa.sin_port = htons(cd->port);
    sa.sin_addr.s_addr = saddr->s_addr;

    if(( n = IO_CONNECT_FUNC(new->fd, (struct sockaddr*)&sa, sizeof sa)) < 0)
    {
      if (errno != EINPROGRESS)
      {
        if(cd->cb != NULL)
	  (*(io_std_cb)cd->cb)(new, IO_CLOSED, cd->cb_arg);
        cd->connected = -1;
	
        if(new->fd > 0)
	  close(new->fd);
        pool_free(p);
	return (void*)1;
      }
    }

    log_debug(ZONE,"Done connect");

    if (n != 0)
    {
      pfdset[0].fd = new->fd;
      pfdset[0].events = ( POLLIN | POLLOUT | POLLHUP | POLLERR );
      pfdset[0].revents = 0;


      if ((n = poll((struct pollfd *)pfdset, 1, 10000)) == 0)  {
	ret = -1;
      }
      else {
	if ((pfdset[0].revents & POLLIN) || (pfdset[0].revents & POLLOUT))  {
	  s=sizeof(error);
	  if (getsockopt( new->fd, SOL_SOCKET, SO_ERROR, &error, &s) < 0)
	    ret = -1;
	}
	else {
	  ret = -1;
	}
      }
      
      if (ret != 0 || error)  {
        if(cd->cb != NULL)
	  (*(io_std_cb)cd->cb)(new, IO_CLOSED, cd->cb_arg);
        cd->connected = -1;
	
        if(new->fd > 0)
	  close(new->fd);
        pool_free(p);
	return (void*)1;
      }
    }

    /* no decrement for servers */
    io_karma(new, KARMA_DEF_INIT, KARMA_DEF_MAX, KARMA_DEF_INC, 0, KARMA_DEF_PENALTY, KARMA_DEF_RESTORE);    

    pthread_mutex_init(&(new->sem),NULL);

    cd->connected = 1; 

    /* set read and write bufor to 32768 */
    s = sizeof(int);
    setsockopt(new->fd,SOL_SOCKET,SO_SNDBUF,(char *)&socket_size,s);
    setsockopt(new->fd,SOL_SOCKET,SO_RCVBUF,(char *)&socket_size,s);
    socket_size = 1;
    setsockopt(new->fd,SOL_SOCKET,SO_KEEPALIVE,(char *)&socket_size,s);

    if(new->cb != NULL)
      (*(io_std_cb)new->cb)(new, IO_NEW, new->cb_arg);
    
    return 0;
}

/*
 * server read 
 */ 
int handle_server_read()
{
  int len;
  char buf[32768]; 

  if (!io__data->serverio) return 0;
  if (io__data->serverio->state != state_ACTIVE) return 0;

  len = IO_READ_FUNC(io__data->serverio->fd, buf, 32761);

  /* if we had a bad read */
  if(len == 0)
  { 
      log_alert(ZONE,"serwer error reading");
      io_close(io__data->serverio);
      return -1; 
  }
  else if(len < 0)
    {
      if(errno != EWOULDBLOCK && errno != EINTR && errno != EAGAIN) 
	{
	  /* kill this socket and move on */
	  log_alert(ZONE,"serwer error reading.");
	  io_close(io__data->serverio);
	  return -1;  /* loop on the same socket to kill it for real */
	}
    }
  else 
    {
      if (len == 32761)
	io__data->serverio->is_data = 1; /* mark data in server socket */
      else
	io__data->serverio->is_data = 0;
   

      buf[len] = '\0';
      log_debug(ZONE, "IO read from serwer socket %d: %s", io__data->serverio->fd, buf);

      /* parse data */
      _io_xml_parser(io__data->serverio, buf, len);
    }
  return 0;
}

int handle_client_read(io cur)
{
  int len, maxlen;
  char buf[8192];

  /* how match can read */
  maxlen = KARMA_READ_MAX(cur->k.val);

  if (maxlen > 8192) maxlen = 8191;
		
  len = IO_READ_FUNC(cur->fd, buf, maxlen);
		
  /* if we had a bad read */
  if(len == 0)
    { 
      log_alert(ZONE,"error in reading");
      io_close(cur);		
      return -1;  /* go */
    }
  else if(len < 0)
    {
      if(errno != EWOULDBLOCK && errno != EINTR && errno != EAGAIN) 
	{
	  /* kill this socket and move on */
	  log_alert(ZONE,"error in reading.");
	  io_close(cur);
	  return -1;  /* go */
	}
    }
  else 
    {      
      /* add to sockets with data */
      if (len < maxlen) {
	if (cur->is_data) {
	  _io_unlink_data(cur);
	  cur->is_data = 0; /* no more data */
	}
      }
      else {
	if (!cur->is_data) {
	  _io_link_data(cur);
	  cur->is_data = 1; /* still data in socket */
	}
      }

      if (karma_check_lk(&cur->k,len))
	{	  
	  if (cur->k.val <= 0)  {
	    log_debug("READ","socket from %s out of karma",cur->ip);
	  }		    
	}
      
      buf[len] = '\0';
      log_debug(ZONE, "IO read from socket %d: %s", cur->fd, buf);
      
      /* parse data */
      _io_xml_parser(cur, buf, len);

      if (cur->state == state_CLOSE){
	return -1;
      }

    }
  return 0;
} 

/* 
 * main select loop thread 
 */
void * io_main(void *arg)
{
    io          cur,
                temp;    
    int         i,
                nfds;
    DWORD       time,
                iotime,
                datatime;
    int length;
    BOOL        iocheck = 0, /* ping */
                datacheck = 0;
    int         server_read_count; /* reads every DEF_READ_SERVER in client loop */
    int free_size=-1;

    /* dev/epoll */
    struct pollfd *pfds;
    struct evpoll evp;

    datatime = iotime = timeGetTimeSec(); /* get sec */

    log_debug(ZONE, "IO is starting up");

    update_links_count();

    /* main loop */
    while (1)
    {
        /* if we are closing down, exit the loop */
        if (io__data->shutdown == 1)
            break;

	/* wait */
	Sleep(1);

	/* get loop time */
	time = timeGetTimeSec();

	/* how often check sockets, that might have data 1s */
	if ( (time-datatime) > wpj->cfg->datacheck )
	{
	  //	  log_debug(ZONE,"data check...");
	  datatime = time;
	  datacheck = 1;
	}

	/* how often check ping */
	if ( (time-iotime) > wpj->cfg->iocheck )
	{
	  log_debug(ZONE,"ping check...");
	  iotime = time;
	  iocheck = 1;
	}

	/* ################# SERVER ######################## */
	/* if server */
	if (io__data->serverio){
	  cur = io__data->serverio;

	  /* if server needs to be closed */
	  if (cur->state == state_CLOSE) {
    
	    fdop(cur->fd, REMFD);

	    _io_close_server(cur);

	    /* set refresh flag */
	    io__data->refresh_flag  = 1;
	    io__data->refresh_start = time;
	    continue;
	  }
	  else
	    /* if init server socket */
	    if ( cur->k.val == KARMA_DEF_INIT )	{
	      io__data->refresh_flag = 0;

	      /* reset the karma to restore val */
	      cur->k.val = cur->k.restore;
		  
	      /* and make sure that they are in the read set */
	      if ( fdop(cur->fd, ADDFD) < 0) {
		log_alert(ZONE,"unable to add client to dev/epoll");
	      }

	      /* try to read from opened server socket */
	      handle_server_read();
	      cur->last_io = time;
	      cur->ping_in_progress = 0;
	      
	      log_debug(ZONE,"add server to dev");
	    }	  
	  /* if server has data for us */
	    else if (cur->is_data){
	      handle_server_read();
	      cur->last_io = time;
	      cur->ping_in_progress = 0;
	    }

	  /* ************ SERVER PING ********** */
	  if (time > cur->last_io + 20) {
	    /* check ping */

	    if (cur->ping_in_progress == 1){
	      /* check bufor */
	      free_size=-1;
	      if (ioctl(cur->fd, TIOCOUTQ, &free_size) == 0) {
		//free_size -> not send data
		log_debug(ZONE,"server ping  bufor %d last bufor %d",
			  free_size,cur->write_buffer_count);
		    
		/* if socket bufor has changed */
		if (free_size != cur->write_buffer_count){
		  log_alert("ping server","server change buf %d last %d",free_size,
			    cur->write_buffer_count);
		  cur->last_io = time;
		  cur->ping_in_progress = 0;
		  continue;
		}
		    
		/* if we need to disconnect */
		if (time > cur->last_io + 30){
		  log_alert("ping server","disconnect server buf %d last %d",free_size,
			    cur->write_buffer_count);
		  io_close(cur);
		  continue;
		}

	      }
	      else {
		/* error ioctl */
		io_close(cur);
		continue;
	      }
	    }
	    /* no ping in progress */
	    else {
		  /* if something to write in queue */
		  if (cur->queue){
		    log_debug("ping server","send ping queue");
		    i = _io_write_dump_server(cur,&length);
		    if (i < 0) {
		      io_close(cur);
		      continue;
		    }
		  }
		  else {
		    /* send ping */
		    log_debug("ping server","send ping spacja");
		    write(cur->fd," ",1);
		  }
		  
		  /* remember bufor size */
		  free_size = -1;
		  if (ioctl(cur->fd, TIOCOUTQ, &free_size) == 0) {
		      
		    if (free_size > 0) {
		      cur->write_buffer_count = free_size;
		      cur->ping_in_progress = 1;
		    }
		    else {
		      cur->ping_in_progress = 0;
		      cur->last_io = time;
		    }
		    
		    log_alert("ping server","bufor size %d",free_size);
		    
		  }
		  else {
		    log_alert("ping server","ioctl error");
		    io_close(cur);
		    continue;
		  }
	    }
	    
	  }/* server ping */
	  
	}else
	  if (io__data->refresh_flag == 1){
	    /* 5 minut to refresh sessions */
	    if (time > io__data->refresh_start + wpj->cfg->reconnect_time)
	      {
		for(cur = io__data->master__list; cur != NULL;) {
		  if (wpj->clients[cur->pollpos].state == 
		      state_REFRESHING_SESSION) {
		    /* close */
		    temp = cur;
		    cur = cur->next;
		    close_client(temp);
		    continue;	      
		  }
		  cur = cur->next;
		}				
		
	      }
	  }
	/* serverio */
	
	
	/* ################# CLIENTS ######################## */	  
	/* loop all sockets, update ping */	
	if (iocheck)
	  for(cur = io__data->master__list; cur != NULL;) {
	    /* check socket that needs to be close */
	    if (cur->state == state_CLOSE){
	      log_debug(ZONE,"not closed socket");
	      temp = cur;
	      cur = cur->next;
	      close_client(temp);
	      continue;
	    }

	    /* check state */
	    /* not authed */
	    if (wpj->clients[cur->pollpos].state == state_UNKNOWN){
	      if (time > wpj->clients[cur->pollpos].time + wpj->cfg->auth_time){
		log_alert(ZONE,"auth time exceeded");
		temp = cur;
		cur = cur->next;
		close_client(temp);
		continue;
	      }
	    }
	    else{
	      /* status AUTHED or REFRESH */
	      /* if time to ping, 3 minutes after lastio */
	      if (time > cur->last_io + wpj->cfg->ping_time){
		log_debug(ZONE,"ping check in socket");
		/* if ping was sent */
		if (cur->ping_in_progress == 1){
		  /* check bufor */
		  free_size=-1;
		  if (ioctl(cur->fd, TIOCOUTQ, &free_size) == 0) {
		    log_debug(ZONE,"ping check in progress bufor %d last bufor %d",
			      free_size,cur->write_buffer_count);
		    
		    /* if socket bufor has changed */
		    if (free_size != cur->write_buffer_count){
		      cur->last_io = time;
		      cur->ping_in_progress = 0;
		      continue;
		    }
		    
		    /* if we need to disconnect */
		    if (time > cur->last_io + wpj->cfg->ioclose){
		      log_alert("ping","disconnect");
		      temp = cur;
		      cur = cur->next;
		      close_client(temp);
		      continue;
		    }
		  }
		  else {
		    temp = cur;
		    cur = cur->next;
		    close_client(temp);
		    continue;
		  }
		}
		else {
		  cur->ping_in_progress = 1;
		  /* if something to write in queue */
		  if (cur->queue){
		    log_debug(ZONE,"send ping queue");
		    _io_write_dump_client(cur);
		  }
		  else {
		    /* send ping */
		    log_debug(ZONE,"send ping spacja");
		    write(cur->fd," ",1);
		  }

		  /* remember bufor size */
		  {
		    free_size=-1;
		    if (ioctl(cur->fd, TIOCOUTQ, &free_size) == 0){
		      
		      if (free_size > 0)
			cur->write_buffer_count = free_size;
		      else{
			cur->ping_in_progress = 0;
			cur->last_io = time;
		      }
		      
		      log_debug(ZONE,"bufor size %d",free_size);
		      
		    }
		    else{
		      log_alert("ioctl","error");
		      temp = cur;
		      cur = cur->next;
		      close_client(temp);
		      continue;
		    }
		  }
		}
	      }

	    }
	    
	    
	    cur = cur->next;
	  }
	
	if (iocheck == 1) iocheck = 0;


	/* loop sockets that might hava data */
	/* here are socket - out of karma in 90% */
	/* check once 1 sek */
	if (datacheck) {
	  
	  /* stats file */
	  update_links_count();

	  /* if not overloded check sockets */
	  if (!wpj->overload)
	  for (cur = io__data->data__list; cur != NULL;) {
	    log_debug(ZONE,"socket %d in data__list",cur->fd);
	    if (karma_increment_lk(&cur->k,time)){ /* if karma > 0 */
	      if (handle_client_read(cur) == -1) {
		temp = cur;
		cur = cur->next;
		close_client(temp);
		continue;
	      }
	      else {
		cur->last_io = time;
		cur->ping_in_progress = 0;
	      }
	    }
	    cur = cur->next;
	  }
	}
	
	if (datacheck == 1) datacheck = 0;	

	/* ################# ACTIVE READ ######################## */
	/* dev/epoll */
	evp.ep_timeout = 0; /* no wait */
	evp.ep_resoff = 0; 
	nfds = ioctl(io__data->epollfd, EP_POLL, &evp);
	pfds = (struct pollfd *) (io__data->epollmap + evp.ep_resoff);
	
	/* check sockets */
	server_read_count = 0; /* how often check server for read */
	for (i = 0; i < nfds; i++) {

	  /* check if we need to read from server */
	  if ((io__data->serverio)&&(pfds[i].fd == io__data->serverio->fd )) {
	    if (pfds[i].revents & POLLIN) {
	      /* read from server */
	      server_read_count = 0;
	      handle_server_read();
	      io__data->serverio->last_io = time;
	      io__data->serverio->ping_in_progress =0;
	    }
	    else
	      /* error server */
	      if (pfds[i].revents & (POLLHUP|POLLERR)) {
		io_close(io__data->serverio);
	      }/* error server */
	  }
	  else
	    if ((io__data->listenio)&&(pfds[i].fd == io__data->listenio->fd )) {
	      /* accept new connection */
	      cur = NULL;
	      while (cur != ((void *)-1)){

		cur = _io_accept(io__data->listenio);

		if (cur != ((void *)-1)) {/* if more sockets */
		  if (cur != NULL) { /* if accepted */
		    /* set karma */
		    cur->k.val = cur->k.restore;
		    log_debug(ZONE,"accepted socket karma restore =>%d",cur->k.restore);
		    
		    /* add do dev/epoll */
		    fdop(cur->fd,ADDFD);
		    add_client(cur->fd,cur);
		    
		    /* update last io time */
		    cur->last_io = time;
		    cur->ping_in_progress = 0;
		    
		    /* time of connection */
		    wpj->clients[cur->pollpos].time = time;
		    
		    /* try to read */
		    if (wpj->overload == 2){
		      _io_link_data(cur);
		      cur->is_data = 1; /* still data in socket */
		    }
		    else
		    if (handle_client_read(cur) == -1) {
		      log_debug(ZONE,"socket close");
		      temp = cur;
		      close_client(temp);
		      continue;
		    }
		    
		    log_debug(ZONE,"accepted socket");
		  } else {
		    log_alert(ZONE,"didn't accept socket");
		  }
		}
	      }
	    }
	    else {
	      server_read_count++;
	      if (server_read_count == DEF_READ_SERVER)  {
		server_read_count = 0;
		handle_server_read(); 
		io__data->serverio->last_io = time;
		io__data->serverio->ping_in_progress =0;
	      }	  

	      /* find cur in client hash */
	      cur = find_client(pfds[i].fd);
	      
	      if (cur == NULL) continue; /* go */

	      if (cur->state == state_CLOSE){
		  temp = cur;
		  close_client(temp);
		  continue;
	      }
    
	      /* read event */
	      if (pfds[i].revents & POLLIN) {
		/* can read */
		/* update last socket reaction */
		cur->last_io = time;
		cur->ping_in_progress = 0;		

		/* do not read if overload */
		if (wpj->overload == 2){
		  if (!cur->is_data){ 
		    _io_link_data(cur);
		    cur->is_data = 1;
		  }
		  continue;
		}

		if (cur->k.val <= 0) {
		  if (!cur->is_data){ 
		    _io_link_data(cur);
		    cur->is_data = 1;
		  }
		  continue;
		}
		
		karma_increment_lk(&cur->k,time);
		
		log_debug(ZONE,"event read karma %d",cur->k.val);

		if (handle_client_read(cur) == -1) {
		  temp = cur;
		  close_client(temp);
		  continue;
		}
		
	      }/*read*/
	      else
		/* write */
		if (pfds[i].revents & POLLOUT) {
		  if (cur->queue != NULL){ /* if something to write */
		    int ret;

		    /* write the current buffer */
		    ret = _io_write_dump_client(cur);
		    if (ret == -1) { /* error */
		      io_close(cur);
		      temp = cur;
		      close_client(temp);
		      continue;
		    }
		  }
		}/* write */
		else
		  /* error */
		  if (pfds[i].revents & (POLLHUP|POLLERR)) {
		    log_alert(ZONE,"socket close - HUB,ERR");
		    io_close(cur);
		    temp = cur;
		    close_client(temp);
		  }/* error */
	    }
	}/*for*/
	
    }/* while(1) end */
    
    return 0;

}

/***************************************************\
*      E X T E R N A L   F U N C T I O N S          *
\***************************************************/

/* 
   starts the io_main() loop
*/
void io_init(void)
{
  int i;
    if(io__data == NULL)
    {
        /* malloc our instance object */
        io__data              = pmalloco(wpj->p, sizeof(_ios));
        io__data->karma_time  = wpj->cfg->karma_time;      

	io__data->listenio    = NULL;
	io__data->serverio    = NULL;

	for (i = 0 ; i < CLIENT_HASH_SIZE ; i++)
	  client_hash[i] = NULL;

	/* init rate */
	RateInit(DEF_RATE_CONN,DEF_RATE_TIME,DEF_CONN);


        /* init /dev/epoll */
        io__data->epollfd = open("/dev/epoll",O_RDWR);
	
        ioctl(io__data->epollfd, EP_ALLOC,wpj->cfg->max_clients+1000);
	io__data->epollmap=(char *)mmap(NULL, EP_MAP_SIZE(wpj->cfg->max_clients+1000), PROT_READ | PROT_WRITE, MAP_PRIVATE, io__data->epollfd, 0);

        log_debug(ZONE,"epoll %d, %p",io__data->epollfd,io__data->epollmap);
	
	/* init working threads */
	pthread_create(&io__data->t, NULL, io_main, NULL);
	pthread_create(&io__data->wt, NULL, write_queue_main, NULL);

    }
}

/*
 * Cleanup function when server is shutting down, closes
 * all sockets, so that everything can be cleaned up
 * properly.
 */
void io_stop(void)
{
    io cur,temp;
    void * ret;

    log_alert(ZONE, "IO is shutting down");

    /* no need to do anything if io__data hasn't been used yet */
    if(io__data == NULL) 
        return;

    /* flag that it is okay to exit the loop */
    io__data->shutdown = 1;
    pthread_join(io__data->t,&ret);

    /* close listen socket */
    if (io__data->listenio)
      _io_close_listen(io__data->listenio);
    
    io__data->listenio = NULL;

    /* loop client socket, and close it */
    for(cur = io__data->master__list; cur != NULL; )
      {       
	log_alert(ZONE,"closing client sockets. data %d",cur->queue_long);
	temp = cur;
	cur = cur->next;
	close_client(temp);
      }
    

    log_alert(ZONE,"IO stop: stopped clients");

    /* end server writing thread */
    io__data->shutdown = 3;
    pthread_join(io__data->wt,&ret);

    /* end connection to server */
    if (io__data->serverio)
      _io_close_server(io__data->serverio);

    io__data->serverio = NULL;

    /* stop rate */
    RateStop();
    
    /* stop dev/epoll */
    munmap(io__data->epollmap, EP_MAP_SIZE(wpj->cfg->max_clients+1000));
    close(io__data->epollfd);

    /* stats file */
    update_links_count();

    log_alert(ZONE,"IO stop end");

    io__data = NULL;
}

/* 
   creates a new io object from a file descriptor
*/
io io_new(int fd, void *cb, void *arg)
{
    io    new    =  NULL;
    pool  p      =  NULL;
    int   flags  =  0;

    if(fd <= 0) 
        return NULL;
    
    /* create the new IO object */
    p           = pool_new();
    new         = pmalloco(p, sizeof(_io));
    new->p      = p;
    new->type   = type_NORMAL;
    new->state  = state_ACTIVE;
    new->fd     = fd;
    new->cb     = (void*)cb;
    new->cb_arg = arg;
    new->pollpos= -1;

    /* set the default karma values */
    io_karma(new, KARMA_DEF_INIT, KARMA_DEF_MAX, KARMA_DEF_INC, KARMA_DEF_DEC, KARMA_DEF_PENALTY, KARMA_DEF_RESTORE);


    /* set the socket to non-blocking */
    flags =  fcntl(fd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    fcntl(fd, F_SETFL, flags);

    /* add to the select loop */
    _io_link(new);

    return new;
}

/* 
 * client call to close the socket 
 */
void io_close(io m) 
{
    if(m == NULL) 
        return;

    m->state = state_CLOSE;
}

/* 
 * writes a str, or xmlnode to the client socket 
 */
void io_write(io m, xmlnode x, char *buffer, int len)
{
    io_wbq new;
    pool p;
    int ret;

    log_debug(ZONE, "io_write called on x: %X buffer: %.*s", x, len, buffer);

    if(m == NULL)    {
      if (x != NULL) xmlnode_free(x);
      log_alert(ZONE,"Writing to non existing socket !!!");
      return;
    }

    /* if there is nothing to write */
    if(x == NULL && buffer == NULL)
        return;

    if ( m->type == type_SERVER )
      pthread_mutex_lock(&(m->sem));

    /* create the pool for this wbq */
    if(x != NULL)
        p = xmlnode_pool(x);
    else
        p = pool_new();

    /* create the wbq */
    new    = pmalloco(p, sizeof(_io_wbq));
    new->p = p;

    /* set the queue item type */
    if(buffer != NULL)  {
        new->type = queue_CDATA;

        if (len == -1)
            len = strlen(buffer);

        /* XXX more hackish code to print the stream header right on a NUL xmlnode socket */
        if(m->type == type_NUL && strncmp(buffer,"<stream:stream",14))  {
            len++;
            new->data = pmalloco(p,len);
            sprintf(new->data,"%.*s/>",len-2,buffer);
        }
	else {
            new->data = pmalloco(p,len);
            memcpy(new->data,buffer,len);
        }
    }
    else {
        new->type = queue_XMLNODE;
        new->data = xmlnode2str(x);
        len = new->data ? strlen(new->data) : 0;
    }

    /* include the \0 if we're special */
    if(m->type == type_NUL) {
        len++;
    }

    /* assign values */
    new->x    = x;
    new->cur  = new->data;

    new->len = len;

    /* put at end of queue */
    if(m->tail == NULL)
        m->queue = new;
    else
        m->tail->next = new;
    m->tail = new;

    m->queue_long++;

    /*----- only for server ----*/
    if ( m->type == type_SERVER )   {
      if (!wpj->overload)   {
	if (m->queue_long > OVERLOAD_NORMAL)  {
	    wpj->overload = 1;
	}
      }
      else
      if (wpj->overload == 1)  {
      	if (m->queue_long > OVERLOAD_CRITICAL)  {
	    wpj->overload = 2;
	    log_alert("--","2");
	}      
      }
      pthread_mutex_unlock(&(m->sem));
    }
    else {
	/* client - always try to write , optymize this !!!! */
      ret = _io_write_dump_client(m);
      if (ret == -1) {     /* error */
	io_close(m);
      }
      
    }
}

/*
   sets karma values
*/
void io_karma(io m, int val, int max, int inc, int dec, int penalty, int restore)
{
    if(m == NULL)
       return;

    m->k.val     = val;
    m->k.max     = max;
    m->k.inc     = inc;
    m->k.dec     = dec;
    m->k.penalty = penalty;
    m->k.restore = restore;
}

void io_karma2(io m, struct karma *k)
{
    if(m == NULL)
       return;

    m->k.val     = k->val;
    m->k.max     = k->max;
    m->k.inc     = k->inc;
    m->k.dec     = k->dec;
    m->k.penalty = k->penalty;
    m->k.restore = k->restore;
}


/*
   pops the last xmlnode from the queue 
*/
xmlnode io_cleanup(io m)
{
    io_wbq     cur;
    
    if(m == NULL || m->queue == NULL) 
        return NULL;


    if ( m->type == type_SERVER )
      pthread_mutex_lock(&(m->sem));

    /* find the first queue item with a xmlnode attached */
    for(cur = m->queue; cur != NULL;)
    {

        /* move the queue up */
        m->queue = cur->next;

        /* set the tail pointer if needed */
        if(m->queue == NULL)
            m->tail = NULL;

        /* if there is no node attached */
        if(cur->x == NULL)
        {
            /* just kill this item, and move on..
             * only pop xmlnodes 
             */
            io_wbq next = m->queue;
            pool_free(cur->p);
            cur = next;
            continue;
        }

        /* and pop this xmlnode */
	if ( m->type == type_SERVER )
	  pthread_mutex_unlock(&(m->sem));

        return cur->x;
    }

    /* no xmlnodes found */
    if ( m->type == type_SERVER )
      pthread_mutex_unlock(&(m->sem));
    return NULL;
}

/* 
 * request to connect to a remote host 
 */

void io_connect(char *host, int port, void *cb, void *cb_arg, int timeout)
{
    connect_data cd = NULL;
    pool         p  = NULL;
    pthread_attr_t attr;


    if(host == NULL || port == 0) 
      return;

    if (io__data->shutdown > 0)
      return;

    p           = pool_new();
    cd          = pmalloco(p, sizeof(_connect_data));
    cd->p       = p;
    cd->ip      = pstrdup(p, host);
    cd->port    = port;
    cd->cb      = cb;
    cd->cb_arg  = cb_arg;
    cd->timeout = timeout;

    /* set detach able */
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
    pthread_create(&cd->t, &attr, _io_connect,  (void*)cd);
    pthread_attr_destroy(&attr);
}


/* 
 * call to start listening with select 
 */
io io_listen(int port, char *listen_host, void *cb, void *arg)
{
    io         new;
    int        fd;
    pool       p      =  NULL;
    int        flags  =  0;

    if (io__data->shutdown > 0)
      return NULL;

    if (io__data->listenio)
      return NULL;

    log_debug(ZONE, "io_select to listen on %d [%s]",port, listen_host);

    /* attempt to open a listening socket */    
    fd = make_netsocket(port, listen_host, NETSOCKET_SERVER);

    /* if we got a bad fd we can't listen */
    if(fd < 0)
    {
        log_alert(NULL, "io_select unable to create listen socket on %d [%s] - %s", port, listen_host, strerror(errno));
        return NULL;
    }

    /* start listening with a max accept queue of 10 */
    if(listen(fd, 10) < 0)
    {
        log_alert(NULL, "io_select unable to listen on %d [%s] - %s", port, listen_host, strerror(errno));
	wpj_shutdown = 1;
        return NULL;
    }

    /* create the new IO object */
    p           = pool_new();
    new         = pmalloco(p, sizeof(_io));
    new->p      = p;
    new->type   = type_LISTEN;
    new->state  = state_ACTIVE;
    new->fd     = fd;
    new->cb     = cb;
    new->cb_arg = arg;
    new->ip     = pstrdup(new->p, listen_host);
    new->pollpos= 0;

    /* set the default karma values */
    io_karma(new, KARMA_DEF_INIT, KARMA_DEF_MAX, KARMA_DEF_INC, KARMA_DEF_DEC, KARMA_DEF_PENALTY, KARMA_DEF_RESTORE);

   
    /* set the socket to non-blocking */
    flags =  fcntl(fd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    fcntl(fd, F_SETFL, flags);

    /* listenio */    
    io__data->listenio = new;

    /* add to dev/epoll */
    if ( fdop(fd, ADDFD) < 0) {
      _io_close_listen(new);
      log_alert(ZONE,"unable to add listen do dev/epoll");
      new = NULL;
    }

    log_debug(ZONE, "io_select starting to listen on %d [%s]", port, listen_host);

    return new;
}
